package mods.immibis.redlogic.wires;

import java.util.Arrays;

import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraftforge.common.ForgeDirection;
import mods.immibis.redlogic.CommandDebug;
import mods.immibis.redlogic.IBundledEmittingTile;
import mods.immibis.redlogic.IBundledUpdatableTile;
import mods.immibis.redlogic.IRedstoneUpdatableTile;

public class BundledTile extends WireTile implements IRedstoneUpdatableTile, IBundledEmittingTile, IBundledUpdatableTile {
	
	private byte[] strength = new byte[16];
	
	@Override
	public void writeToNBT(NBTTagCompound tag) {
		super.writeToNBT(tag);
		
		tag.setByteArray("strength", strength);
	}
	
	@Override
	public void readFromNBT(NBTTagCompound tag) {
		super.readFromNBT(tag);
		
		strength = tag.getByteArray("strength");
	}
	
	@Override
	protected boolean canConnectToWire(WireTile wire) {
		return super.canConnectToWire(wire) || wire instanceof InsulatedRedAlloyTile;
	}

	@Override
	public boolean canUpdate() {
		return false;
	}
	
	private byte[] oldStrength = new byte[16];
	
	private boolean isUpdating;
	private boolean recursiveUpdatePending = true;
	
	private void updateStrengthFromSurroundingBlocks() {
		{
			byte[] temp = oldStrength;
			oldStrength = strength;
			strength = temp;
			Arrays.fill(strength, (byte)0);
		}
		
		for(int side = 0; side < 6; side++) {
			for(int dir = 0; dir < 6; dir++) {
				if(connectsInDirection(side, dir)) {
					ForgeDirection fd = ForgeDirection.VALID_DIRECTIONS[dir];
					int x = xCoord + fd.offsetX, y = yCoord + fd.offsetY, z = zCoord + fd.offsetZ;
					
					if(connectsInDirectionAroundCorner(side, dir)) {
						fd = ForgeDirection.VALID_DIRECTIONS[side];
						x += fd.offsetX;
						y += fd.offsetY;
						z += fd.offsetZ;
					}
					
					TileEntity te = worldObj.getBlockTileEntity(x, y, z);
					if(te instanceof InsulatedRedAlloyTile) {
						InsulatedRedAlloyTile o = (InsulatedRedAlloyTile)te;
						int o_strength = o.getRedstoneSignalStrength() - 1;
						int colour = o.getType().ordinal() - EnumWireType.INSULATED_0.ordinal();
						if((strength[colour] & 0xFF) < o_strength)
							strength[colour] = (byte)o_strength;
						
					} else if(te instanceof IBundledEmittingTile) {
						byte[] o_strength = ((IBundledEmittingTile)te).getBundledCableStrength();
						for(int k = 0; k < 16; k++) {
							int o_c_strength = (o_strength[k] & 0xFF) - 1;
							if((strength[k] & 0xFF) < o_c_strength)
								strength[k] = (byte)o_c_strength;
						}
					}
				}
			}
		}
	}
	
	private void updateStrength() {
		if(worldObj.isRemote)
			return;
		
		if(isUpdating) {
			recursiveUpdatePending = true;
			return;
		}
		
		if(CommandDebug.WIRE_DEBUG_PARTICLES)
			debugEffect_bonemeal();
		
		isUpdating = true;
		
		do {
			recursiveUpdatePending = false;
			
			updateStrengthFromSurroundingBlocks();
			
			if(!Arrays.equals(oldStrength, strength)) {
				boolean any_decreased = false;
				for(int k = 0; k < 16; k++) {
					if((strength[k] & 0xFF) < (oldStrength[k] & 0xFF)) {
						strength[k] = 0;
						any_decreased = true;
					}
				}
				
				if(any_decreased) {
					updateConnectedThings();
					updateStrengthFromSurroundingBlocks();
				}
				
				updateConnectedThings();
			}
			
		} while(recursiveUpdatePending);
		
		isUpdating = false;
	}
	
	private void updateConnectedThings() {
		int notifiedSides = 0;
		
		if(CommandDebug.WIRE_DEBUG_PARTICLES)
			debugEffect_bonemeal();
		
		for(int side = 0; side < 6; side++) {
			for(int dir = 0; dir < 6; dir++) {
				if(connectsInDirection(side, dir)) {
					ForgeDirection fd = ForgeDirection.VALID_DIRECTIONS[dir];
					int x = xCoord + fd.offsetX, y = yCoord + fd.offsetY, z = zCoord + fd.offsetZ;
					
					if(connectsInDirectionAroundCorner(side, dir)) {
						fd = ForgeDirection.VALID_DIRECTIONS[side];
						x += fd.offsetX;
						y += fd.offsetY;
						z += fd.offsetZ;
					
					} else {
						if((notifiedSides & (1 << dir)) != 0)
							continue;
						notifiedSides |= 1 << dir;
					}
					
					TileEntity t = worldObj.getBlockTileEntity(x, y, z);
					if(t instanceof IBundledUpdatableTile) {
						((IBundledUpdatableTile)t).onBundledInputChanged();
					}
				}
			}
		}
	}

	@Override
	public void onRedstoneInputChanged() {
		updateStrength();
	}
	
	@Override
	public void onBundledInputChanged() {
		updateStrength();
	}
	
	@Override
	public byte[] getBundledCableStrength() {
		return strength;
	}
	
	@Override
	protected boolean debug(EntityPlayer ply) {
		if(!worldObj.isRemote)
		{
			int[] i = new int[16];
			for(int k = 0; k < 16; k++)
				i[k] = strength[k] & 0xFF;
			ply.sendChatToPlayer("Bundled cable strength: " + Arrays.toString(i));
		}
		return true;
	}

}
